----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    13:27:42 06/14/2011 
-- Design Name: 
-- Module Name:    Sigma delta 1st order Modulator 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.std_logic_arith.all;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity SigmaDelta is
   generic(
      WBUFFER: natural := 14 -- Buffer depth X16  14 -> 16k ok
   );
	port (
		CLK1: IN STD_LOGIC;				            -- clock (naturally)
   	CLK2: IN STD_LOGIC;				            -- clock (naturally) not used
		btn: IN STD_LOGIC_VECTOR (3 DOWNTO 0);	   -- Buttons on Basys
		seg: OUT STD_LOGIC_VECTOR (6 DOWNTO 0);	-- Segments
		an: OUT STD_LOGIC_VECTOR (3 DOWNTO 0);	   -- Digit
		dp: OUT STD_LOGIC;				               -- dot in Segment
		led: OUT STD_LOGIC_VECTOR (7 DOWNTO 0);	-- Segments
      sw: IN STD_LOGIC_VECTOR (7 DOWNTO 0);
      JA: OUT STD_LOGIC_VECTOR (7 downto 0);      -- Output stream
		JB1: OUT STD_LOGIC_VECTOR (5 downto 0);     --    NEXYS3 
		JB2: In STD_LOGIC_VECTOR (1 downto 0);      --    NEXYS3 
      JC: OUT STD_LOGIC_VECTOR (7 downto 0);      -- Output stream
		TxD: OUT STD_LOGIC;        --    NEXYS3 
		RxD: IN STD_LOGIC          --    NEXYS3 
	);
end SigmaDelta;

-- JB2(0) Analog input
-- JB2(1) DAC output
-- JB1(3) Compare out

-- switch (2 downto 0) control pulsewidth EN_X and oversampling rate EN_T
-- sample rate pulsewidth: 10ns; 100ns, 1us, 10us
-- oversampling: 2048, 128
-- 000  2048, 10ns
-- 001   128, 10ns
-- 010  2048, 240ns
-- 011   128, 240ns
-- 100  2048, 1us
-- 101   128, 1us
-- 110  2048, 10us
-- 111   128, 10us

-- switch (4 downto 3) operation
-- 00  SAR operation
-- 01  sawtooth down DAC
-- 10  sawtooth up DAC
-- 11  sine signal DAC

-- switch (5 downto 6) control number of bits
-- 00  (32)  16
-- 01  (16)  12 
-- 10   (8)  10
-- 11   (2)  8

-- switch (sw7) = 1 controls enabling inverted data for capacitive C2C

architecture Behavioral of SigmaDelta is
 
 -- Component Declaration
    COMPONENT SegCntrl
      PORT(
		            CLK1: IN STD_LOGIC;				            -- clock (naturally)
                Digit0: IN STD_LOGIC_VECTOR (3 DOWNTO 0);
		          Digit1: IN STD_LOGIC_VECTOR (3 DOWNTO 0);
		          Digit2: IN STD_LOGIC_VECTOR (3 DOWNTO 0);
		          Digit3: IN STD_LOGIC_VECTOR (3 DOWNTO 0);
		           reset: IN STD_LOGIC;
		             seg: OUT STD_LOGIC_VECTOR (6 DOWNTO 0);	-- Segments
		              an: OUT STD_LOGIC_VECTOR (3 DOWNTO 0);	   -- Digit
		              dp: OUT STD_LOGIC				               -- dot in Segment
                  );
    END COMPONENT;

    COMPONENT CLK_DIV 
    Port ( CLK : in  STD_LOGIC;         -- 100 MHz ; 10ns
	        reset : in STD_LOGIC;        
           EN_4 : out  STD_LOGIC;       --  25 MHz ;  40ns
           EN_16 : out  STD_LOGIC;      --   6 MHz ; 240ns
           EN_32 : out  STD_LOGIC;      --   3 MHz ; 480ns
           EN_64 : out  STD_LOGIC;      --   1.5 MHz ; 1us
           EN_128 : out  STD_LOGIC;     --   750 kHz ; 2us
           EN_256 : out  STD_LOGIC;     --   375 kHz ;  4 us
           EN_512 : out  STD_LOGIC;     --   187 kHz ;  8 us
           EN_1024 : out  STD_LOGIC;    --    93 kHz ; 16 us
           EN_2048 : out  STD_LOGIC;    --    46 kHz ; 32 us 
           EN_4096 : out  STD_LOGIC;    --    23 kHz ; 64 us
           EN_65K : out  STD_LOGIC;     --    1.25 kHz; 1 ms
           EN_16K : out  STD_LOGIC;     --     5 kHz ; 256 us
           EN_8K : out  STD_LOGIC;      --     11kHz ; 128 us
			  EN_1M : out  STD_LOGIC);     --     75 Hz ; 16 ms
    end COMPONENT;

  COMPONENT uart
   port(
      clk, reset: in std_logic;
      rd_uart, wr_uart: in std_logic;
      rx: in std_logic;
      w_data: in std_logic_vector(7 downto 0);
      tx_full, tx_empty, rx_empty: out std_logic;
      r_data: out std_logic_vector(7 downto 0);
      tx: out std_logic
   );
   end COMPONENT;

	COMPONENT Sine
	PORT(
		COUNT : IN std_logic_vector(8 downto 0);          
		SIN : OUT std_logic_vector(15 downto 0)
		);
	END COMPONENT;

	COMPONENT Sine12
	PORT(
		COUNT : IN std_logic_vector(13 downto 0);          
		SIN : OUT std_logic_vector(15 downto 0)
		);
	END COMPONENT;

	COMPONENT SINC2
   generic(
      B: natural:=24 -- number of bits
   );
	PORT(
		RES : IN std_logic;
		DIN : IN std_logic;
		CLK : IN std_logic;
		FSCLK : IN std_logic;
		FBWCLK : IN std_logic;          
		DOUT : OUT std_logic_vector(B downto 0)
		);
	END COMPONENT;
  
   COMPONENT DPRAM
   generic(
      B: natural := 16; -- number of bits
      W: natural := 11 -- number of address bits
   );
	PORT(
		clk : IN std_logic;
		we : IN std_logic;
		w_data : IN std_logic_vector(B - 1 downto 0);
		w_addr : IN std_logic_vector(W - 1 downto 0);          
		r_addr : IN std_logic_vector(W - 1 downto 0);          
		r_data : OUT std_logic_vector(B - 1 downto 0)
		);
	END COMPONENT;


SIGNAL Code, BCode, BACode, BRCode, CCode: STD_LOGIC_VECTOR (15 downto 0);
SIGNAL DeltaCode: STD_LOGIC_VECTOR (15 downto 0);
SIGNAL C1Code, HCode: STD_LOGIC_VECTOR (15 downto 0);
SIGNAL TCode, T1Code, STCode, SIN: STD_LOGIC_VECTOR (15 downto 0);
SIGNAL SCode, MZero, MOne: STD_LOGIC_VECTOR (17 downto 0);
Signal CZERO: STD_LOGIC_VECTOR (WBUFFER - 1 downto 0);

SIGNAL DOUT : std_logic_vector(26 downto 0);
SIGNAL D, DX, DT: STD_LOGIC;
SIGNAL SSAR: STD_LOGIC; -- conversion finished
SIGNAL stBit: STD_LOGIC; -- start flag for buffer
SIGNAL sinv: STD_LOGIC;
SIGNAL Count16: STD_LOGIC_VECTOR (3 downto 0);
SIGNAL muxs: STD_LOGIC_VECTOR (2 downto 0);
SIGNAL JAX: STD_LOGIC_VECTOR (5 downto 0);
SIGNAL reset: STD_LOGIC;
SIGNAL ENX, EN_4, EN_16, EN_32, EN_64, EN_128, EN_256, EN_4096, EN_8K, EN_16K, EN_65K, EN_1M: STD_LOGIC;
SIGNAL EN_512, EN_1024, EN_2048, EN_T, EN_X: STD_LOGIC;
SIGNAL Sel: STD_LOGIC_VECTOR (2 downto 0);

SIGNAL FULL, we, we1: STD_LOGIC;                                   -- new data available ?
SIGNAL wr_uart,tx_full, tx_empty, rx_empty: STD_LOGIC;
SIGNAL nData: STD_LOGIC;                                   -- new data available ?
SIGNAL w_data,RD_data: STD_LOGIC_VECTOR(7 downto 0);
SIGNAL i_data: STD_LOGIC_VECTOR(4 downto 0);
SIGNAL ST: STD_LOGIC_VECTOR(4 downto 0);
SIGNAL decode: STD_LOGIC_VECTOR(2 downto 0);
SIGNAL trigger: STD_LOGIC_VECTOR(2 downto 0);
SIGNAL Valid: STD_LOGIC;
SIGNAL TXI: STD_LOGIC;
SIGNAL MaxCount: STD_LOGIC_VECTOR(8 downto 0);
SIGNAL douta: STD_LOGIC_VECTOR(11 downto 0);
SIGNAL addra: STD_LOGIC_VECTOR(13 downto 0);
SIGNAL wea: STD_LOGIC_VECTOR(0 downto 0);
SIGNAL dpx:STD_LOGIC;
SIGNAL AVG:STD_LOGIC;


begin

-- Differential input

IBUFDS_inst : IBUFDS
generic map (
   DIFF_TERM => FALSE, -- Differential Termination 
   IBUF_LOW_PWR => TRUE, -- Low power (TRUE) vs. performance (FALSE) setting for referenced I/O standards
   IOSTANDARD => "DEFAULT")
port map (
   O => D,  -- Buffer output
   I => JB2(0),  -- Diff_p buffer input (connect directly to top-level port)
   IB => JB2(1) -- Diff_n buffer input (connect directly to top-level port)
);

CZERO <= ( others => '0');
-- reset
 reset<=btn(0);

-- led
 led <= Code(15 downto 8);

-- CLK Divider Nexys 3 Board 100MHz, 10ns
-- EN_1M 100Hz, 10ms; EN_1024 100kHz, 10us
-- EN_256 50kHz,

 CLKDIV0: CLK_DIV port map (
             CLK => CLK1, reset => reset, EN_4 => EN_4, EN_16 => EN_16, 
				 EN_32 => EN_32, EN_64 => EN_64, EN_128 => EN_128, 
				 EN_256 => EN_256, EN_4096 => EN_4096, EN_2048 => EN_2048, 
				 EN_1024 => EN_1024,EN_512 => EN_512, EN_16K => EN_16K,
             EN_8K => EN_8K,				 
				 EN_65K => EN_65K, EN_1M => EN_1M);

-- 7 Segment Display
 dp <= dpx;
 
 SEG0: Segcntrl port map (
             CLK1 => CLK1, Digit0 => Code(3 downto 0), Digit1 => Code(7 downto 4), 
				 Digit2 => Code(11 downto 8), Digit3 => Code(15 downto 12), 
             reset => reset, seg => seg, an => an, dp => dpx);

-----------------------------------------------------------------------------------
-- What goes to output JA,JB for DAC
--
--  Scode ADC output, BCode counter, SIN sine function
--  SCode,BCode,SIN => C1code => CCode
--
-----------------------------------------------------------------------------------
-- switch (6 downto 5) control number of bits
-- 1st order: SNR = 6.02 N + 1.76 - 5.17 + 30 logK 
--                  N=1   K =  10 ^((B *6 - 2) / 30)  
-- 00  (32)  16   K = 1024
-- 01  (16)  12   K = 256
-- 10   (8)  10   K = 128
-- 11   (2)  8    K = 32
  -- DAC driver
  JA <= CCode(7 downto 0);
  
  -- JA <= CCode(7 downto 0) when sw(6 downto 5)="00" else
  --		  CCode(3 downto 0) & "0000" when sw(6 downto 5)="01" else 
  --		  CCode(1 downto 0) & "000000" when sw(6 downto 5)="10" else 
  --		  "00000000";
  JC <= CCode(15 downto 8);
  -- JC <= CCode(15 downto 8) when sw(6 downto 5)="00" else
  --		  CCode(11 downto 4) when sw(6 downto 5)="01" else 
  --		  CCode(9 downto 2) when sw(6 downto 5)="10" else 
  --		  CCode(7 downto 0);
  -- HEX display
  Code <= HCode(15 downto 0);
  -- Code <= HCode(15 downto 0) when sw(6 downto 5)="00" else
  --		    HCode(11 downto 0) & "0000" when sw(6 downto 5)="01" else 
  -- 		    HCode(9 downto 0) & "000000" when sw(6 downto 5)="10" else 
  --		    HCode(7 downto 0) & "00000000";
  DeltaCode <= "0000000000000001" when sw(6 downto 5)="00" else  -- +1
               "0000000000001001" when sw(6 downto 5)="01" else  -- +9
   		      "0000000000000101" when sw(6 downto 5)="10" else  -- +5
               "0000000000101011"; -- 43
-- switch (4 downto 3) operation
-- 00  SAR operation
-- 01  sawtooth down DAC
-- 10  sawtooth up DAC
-- 11  sine signal DAC

    muxs <= sw(7) & sw(4 downto 3); -- sw(7)='1' capacitance one bit less and '1' at front
    

    C1Code <= '1' & SCode(16 downto 2) when muxs = "100" else
	           SCode(16 downto 1) when muxs = "000" else
	           BACode when muxs = "001" else
	           '1' & BACode(15 downto 1) when muxs = "101" else
	           not(BACode) when muxs = "010" else
	           '1' & not(BACode(15 downto 1)) when muxs = "110" else
	           SIN when muxs = "011" else
	           '1' & SIN(15 downto 1);
    
	 SCode <= "00" & TCode;

-- Display
-----------------------------------------------------------------------------------
-- slow update display
-----------------------------------------------------------------------------------
   process(CLK1,reset,EN_1M) -- previous EN_T
     begin
       if ( reset = '1') then   
           HCode<=(others => '0');
       elsif  ((CLK1'event) and CLK1='1') and EN_1M='1' then 
           HCode <= TCode;
		  end if;
     end process;
    
    

-- Inverted signal for half the cycle when sw(7) = '1'
    CCode <= C1Code;
  

-----------------------------------------------------------------------------------
-- Converter should generate TCode
-- JB1(4) <= for each new value '1'
-----------------------------------------------------------------------------------

	Inst_SINC2: SINC2 
	generic map(B=>26)
   PORT MAP(
		RES => reset,
		DIN => DX,
		CLK => CLK1,
		FSCLK => EN_X,
		FBWCLK => EN_T,
		DOUT => DOUT
	);

	TCode <= DOUT (26 downto 11);
	
	JB1(4) <= EN_T; -- Output clock rate

-----------------------------------------------------------------------------------
-- RAM to buffer values
-- BCode and BRCode should have the same length as RAM address
-----------------------------------------------------------------------------------
	Inst_RAM: DPRAM 
	generic map(
      B => 16, -- number of bits
      W => WBUFFER -- number of address bits 2048 values
   )
	PORT MAP(
		clk => CLK1 ,
		we => we1,
		w_data => TCode ,
		w_addr => BCode(WBUFFER - 1 downto 0),
		r_addr => BRCode(WBUFFER - 1 downto 0),
		r_data => T1Code
	);
   we1 <= we and EN_T; 

-----------------------------------------------------------------------------------
-- Counter Code for sawtooth BCode EN_T
-----------------------------------------------------------------------------------
   process(CLK1,reset,EN_T) -- previous EN_T
     begin
       if ( reset = '1') then   
            BCode<=(others => '0');
            BACode<=(others => '0');
       elsif  ((CLK1'event) and CLK1='1') and EN_T = '1' then 
           BCode <= unsigned(BCode) + 1;   -- 
           BACode <= unsigned(BACode) + unsigned(DeltaCode); -- primzahl 9, 43 eine periode, 1129
       end if;
     end process;
    
-----------------------------------------------------------------------------------
-- Data sampling with EN_X
-----------------------------------------------------------------------------------
   process(CLK1,reset,EN_X) -- previous EN_T
     begin
       if ( reset = '1') then   
				DX <='0';
       elsif  ((CLK1'event) and CLK1='1') and EN_X='1' then 
            DX <= D;			  
       end if;
     end process;
    
-----------------------------------------------------------------------------------
-- New Data out with EN_T
-----------------------------------------------------------------------------------
   process(CLK1,reset,EN_T) -- previous EN_T
     begin
       if ( reset = '1') then   
				DT <='0';
       elsif  ((CLK1'event) and CLK1='1') and EN_T='1' then 
            DT <= not(DT);			  
       end if;
     end process;

-----------------------------------------------------------------------------------
-- Sine signal SIN
-----------------------------------------------------------------------------------
	Inst_Sine: Sine12 PORT MAP(
		COUNT => BACode(13 downto 0), --
		SIN => SIN
	);


-----------------------------------------------------------------------------------
-- Switches control CLK rate of sampling
-- Output EN_X sampling rate
--        EN_T reduced output rate 
--    Port ( CLK : in  STD_LOGIC;         -- 100 MHz ; 10ns
--	        reset : in STD_LOGIC;        
--           EN_4 : out  STD_LOGIC;       --  25 MHz ;  40ns
--           EN_16 : out  STD_LOGIC;      --   6 MHz ; 240ns
--           EN_32 : out  STD_LOGIC;      --   3 MHz ; 480ns
--           EN_64 : out  STD_LOGIC;      --   1.5 MHz ; 1us
--           EN_128 : out  STD_LOGIC;     --   750 kHz ; 2us
--           EN_256 : out  STD_LOGIC;     --   375 kHz ;  4 us
--           EN_512 : out  STD_LOGIC;     --   187 kHz ;  8 us
--           EN_1024 : out  STD_LOGIC;    --    93 kHz ; 16 us
--           EN_2048 : out  STD_LOGIC;    --    46 kHz ; 32 us 
--           EN_4096 : out  STD_LOGIC;    --    23 kHz ; 64 us
--           EN_65K : out  STD_LOGIC;     --    1.25 kHz; 1 ms
--           EN_16K : out  STD_LOGIC;     --     5 kHz ; 256 us
--           EN_8K : out  STD_LOGIC;      --     11kHz ; 128 us
--			  EN_1M : out  STD_LOGIC);     --     75 Hz ; 16 ms
-----------------------------------------------------------------------------------
-- switch (2 downto 0) control pulsewidth EN_X and oversampling rate EN_T
-- sample rate pulsewidth: 10ns; 100ns, 1us, 10us
-- oversampling: 1024.. 16k
-- 000  1024, 10ns
-- 001  2048, 10ns
-- 010  1024, 120ns
-- 011  4048, 240ns
-- 100  2048,   1us
-- 101  16k,    1us
-- 110  1024,  10us
-- 111  4048,  10us

       Sel <= sw(2 downto 0);
		 EN_X <= EN_4 when Sel="000" else -- 16 Times EN_T
               EN_4 when Sel="001" else
               EN_16 when Sel="010" else -- 32 times send
               EN_16 when Sel="011" else -- 8 times
				   -- EN_2048, EN_1024, EN_512
               EN_64 when Sel="100" else -- 4 times
               EN_64 when Sel="101" else -- 2 times
               EN_1024 when Sel="110" else -- 1 time
               EN_256;

		 EN_T <= EN_8K when Sel="000" else  -- 8M x16 needed
               EN_16K when Sel="001" else
               EN_16K when Sel="010" else -- EN_65K x16 4 times send (x32 x8 send)
               EN_65K when Sel="011" else -- 2 times send
               EN_65K when Sel="100" else -- 
               EN_1M when Sel="101" else
               EN_1M when Sel="110" else
               EN_1M;

-----------------------------------------------------------------------------------
-- Start UART
-----------------------------------------------------------------------------------
-- UART implementation
--
--  Input TCode(15..0)
--
-- 16 Bit TCode is transfered to STCode (with ST=0001) to i_data to w_data 
-- ST is counted with EN_16K
-- ST is state machine to 
--
-- Transfer Code
-- Change EN_16K according to sampling rate
-- Counter to send 8 data  x.....<cr><lf>  

   process(CLK1,EN_16K,reset) -- previous EN_16K
     begin
       if ( reset = '1') then   
            ST <="10000";
				FULL <= '1';
            STCode <= (others => '0');				
            BRCode <= (others => '0');				
       elsif  ((CLK1'event) and CLK1='1')  then -- and EN_16K = '1' 
			                         -- and (tx_empty) prevents overflow of sending
											 
           if ((tx_empty='1') and (EN_T ='1') AND (ST ="10000") ) then  
			                         -- ready for next data, get next data to send
               ST <="00000";
               STCode <= T1Code;
               BRCode <= unsigned(BRCode) + 1;
					if (BRCode(WBUFFER-1 downto 0) = CZERO ) then
					  FULL <= not(FULL);
					end if;
			  end if;
			  if (ST < "10000") then  
             ST <= unsigned(ST) + 1;			   
			  end if;
       end if;
     end process;
  
  stBit <= '1' when (BRCode(WBUFFER-1 downto 0) = CZERO ) else 
           '0';
			  
  we <= '1' when ( FULL ='1' ) else
        '0';  -- DRAM write enable is ok

  wr_uart<= '1' when ( (ST(0) = '1') and (FULL = '0')) else
            '0'; -- UART write enable
				
-- Write x<BCode>crlf
  decode <= ST(3 downto 1);
  i_data <= '0' & STCode(3 downto 0) when decode="101" else
            '0' & STCode(7 downto 4) when decode="100" else
				'0' & STCode(11 downto 8) when decode="011" else
				'0' & STCode(15 downto 12) when decode="010" else
				'0' & "0000" when decode="001" else
			--	"10000" when decode="001" else -- ASCII 0
				'1' & stBit & "001" when decode="000" else -- ASCII  x start y
				"10010" when decode="110" else -- ASCII cr
				"10011" when decode="111" else -- ASCII lf
				"10100";

  w_data <= "00110000" when i_data="00000" else   -- HEX 30 ASCII "0"
            "00110001" when i_data="00001" else   -- 1
            "00110010" when i_data="00010" else   -- 2
            "00110011" when i_data="00011" else   -- 3
            "00110100" when i_data="00100" else   -- 4
            "00110101" when i_data="00101" else   -- 5
            "00110110" when i_data="00110" else   -- 6
            "00110111" when i_data="00111" else   -- 7
            "00111000" when i_data="01000" else   -- 8
            "00111001" when i_data="01001" else   -- 9
            "01000001" when i_data="01010" else   -- A
            "01000010" when i_data="01011" else   -- B
            "01000011" when i_data="01100" else   -- C
            "01000100" when i_data="01101" else   -- D
            "01000101" when i_data="01110" else   -- E
            "01000110" when i_data="01111" else   -- F
            "00110000" when i_data="10000" else  -- 0  
            "01111000" when i_data="10001" else  -- x
				"00001101" when i_data="10010" else  -- cr
				"00001010" when i_data="10011" else  -- lf
				"00100000" when i_data="10100" else  -- space
            "01000111" when i_data="11001" else  --  G ,Start of block
				"00100000";

  UART0: uart   port map(
      clk=>CLK1, reset=>reset, rd_uart=>'0', wr_uart=>wr_uart, rx=>RXD,
      w_data => w_data, tx_full=> tx_full, tx_empty => tx_empty, rx_empty=> rx_empty,
      r_data => RD_data, tx=> TXI
   );

-----------------------------------------------------------------------------------
-- End UART
-----------------------------------------------------------------------------------
	
-- Monitor seriell Interface Connnector JB
-- Baud Rate 38400 8 Bit N -1  
-- 16 Bit -> 24 Bit? 1766/s -> 500us (16Bit) -> 30us Maximalfrequenz SAR

	TXD <= TXI;
	JB1(0) <= TXI;
	JB1(1) <= DT;
	JB1(2) <= NOT(DX);
	JB1(3) <= DX;
	JB1(5) <= EN_X;
	
end Behavioral;

